@@ -55,17 +55,30 @@
 //! build`. This is a convenience command so you don't forget any steps. Alternatively, you could
 //! write a Makefile for your project.
 
-use std::path::{Path, PathBuf};
-use std::result;
+#![allow(clippy::let_unit_value)]
+#![warn(
+    elided_lifetimes_in_paths,
+    single_use_lifetimes,
+    clippy::absolute_paths,
+    clippy::wildcard_imports
+)]
+#![deny(unsafe_op_in_unsafe_fn)]
+
+use std::ffi::OsStr;
+use std::ffi::OsString;
+use std::path::Path;
+use std::path::PathBuf;
+
+use anyhow::anyhow;
+use anyhow::Context as _;
+use anyhow::Result;
 
 use tempfile::tempdir;
-use thiserror::Error;
+use tempfile::TempDir;
 
 // libbpf-cargo binary is the primary consumer of the following modules. As such,
 // we do not use all the symbols. Silence any unused code warnings.
 #[allow(dead_code)]
-mod btf;
-#[allow(dead_code)]
 mod build;
 #[allow(dead_code)]
 mod gen;
@@ -77,17 +90,6 @@ mod metadata;
 #[cfg(test)]
 mod test;
 
-/// Canonical error type for this crate.
-#[derive(Error, Debug)]
-pub enum Error {
-    #[error("Error building BPF object file: {0}")]
-    Build(String),
-    #[error("Error generating skeleton: {0}")]
-    Generate(String),
-}
-
-pub type Result<T> = result::Result<T, Error>;
-
 /// `SkeletonBuilder` builds and generates a single skeleton.
 ///
 /// This interface is meant to be used in build scripts.
@@ -97,35 +99,60 @@ pub type Result<T> = result::Result<T, Error>;
 /// ```no_run
 /// use libbpf_cargo::SkeletonBuilder;
 ///
-/// SkeletonBuilder::new("myobject.bpf.c")
+/// SkeletonBuilder::new()
+///     .source("myobject.bpf.c")
 ///     .debug(true)
 ///     .clang("/opt/clang/clang")
-///     .generate("/output/path")
+///     .build_and_generate("/output/path")
 ///     .unwrap();
 /// ```
 pub struct SkeletonBuilder {
     debug: bool,
-    source: PathBuf,
+    source: Option<PathBuf>,
+    obj: Option<PathBuf>,
     clang: Option<PathBuf>,
-    clang_args: String,
+    clang_args: Vec<OsString>,
     skip_clang_version_check: bool,
     rustfmt: PathBuf,
+    dir: Option<TempDir>,
+}
+
+impl Default for SkeletonBuilder {
+    fn default() -> Self {
+        Self::new()
+    }
 }
 
 impl SkeletonBuilder {
-    /// Create a new builder instance, where `source` is the path to the BPF object source
-    /// (typically suffixed by `.bpf.c`)
-    pub fn new<P: AsRef<Path>>(source: P) -> Self {
+    pub fn new() -> Self {
         SkeletonBuilder {
             debug: false,
-            source: source.as_ref().to_path_buf(),
+            source: None,
+            obj: None,
             clang: None,
-            clang_args: String::new(),
+            clang_args: Vec::new(),
             skip_clang_version_check: false,
             rustfmt: "rustfmt".into(),
+            dir: None,
         }
     }
 
+    /// Point the [`SkeletonBuilder`] to a source file for compilation
+    ///
+    /// Default is None
+    pub fn source<P: AsRef<Path>>(&mut self, source: P) -> &mut SkeletonBuilder {
+        self.source = Some(source.as_ref().to_path_buf());
+        self
+    }
+
+    /// Point the [`SkeletonBuilder`] to an object file for generation
+    ///
+    /// Default is None
+    pub fn obj<P: AsRef<Path>>(&mut self, obj: P) -> &mut SkeletonBuilder {
+        self.obj = Some(obj.as_ref().to_path_buf());
+        self
+    }
+
     /// Turn debug output on or off
     ///
     /// Default is off
@@ -136,26 +163,37 @@ impl SkeletonBuilder {
 
     /// Specify which `clang` binary to use
     ///
-    /// Default searchs `$PATH` for `clang`
+    /// Default searches `$PATH` for `clang`
     pub fn clang<P: AsRef<Path>>(&mut self, clang: P) -> &mut SkeletonBuilder {
         self.clang = Some(clang.as_ref().to_path_buf());
         self
     }
 
-    /// Pass additional arguments to `clang` when buildling BPF object file
+    /// Pass additional arguments to `clang` when building BPF object file
     ///
     /// # Examples
     ///
     /// ```no_run
     /// use libbpf_cargo::SkeletonBuilder;
     ///
-    /// SkeletonBuilder::new("myobject.bpf.c")
-    ///     .clang_args("-DMACRO=value -I/some/include/dir")
-    ///     .generate("/output/path")
+    /// SkeletonBuilder::new()
+    ///     .source("myobject.bpf.c")
+    ///     .clang_args([
+    ///         "-DMACRO=value",
+    ///         "-I/some/include/dir",
+    ///     ])
+    ///     .build_and_generate("/output/path")
     ///     .unwrap();
     /// ```
-    pub fn clang_args<S: AsRef<str>>(&mut self, opts: S) -> &mut SkeletonBuilder {
-        self.clang_args = opts.as_ref().to_string();
+    pub fn clang_args<A, S>(&mut self, args: A) -> &mut SkeletonBuilder
+    where
+        A: IntoIterator<Item = S>,
+        S: AsRef<OsStr>,
+    {
+        self.clang_args = args
+            .into_iter()
+            .map(|arg| arg.as_ref().to_os_string())
+            .collect();
         self
     }
 
@@ -175,43 +213,72 @@ impl SkeletonBuilder {
         self
     }
 
-    /// Generate the skeleton at path `output`
-    pub fn generate<P: AsRef<Path>>(&self, output: P) -> Result<()> {
-        let filename = self
+    /// Build BPF programs and generate the skeleton at path `output`
+    pub fn build_and_generate<P: AsRef<Path>>(&mut self, output: P) -> Result<()> {
+        self.build()?;
+        self.generate(output)?;
+
+        Ok(())
+    }
+
+    // Build BPF programs without generating a skeleton.
+    //
+    // [`SkeletonBuilder::source`] must be set for this to succeed.
+    pub fn build(&mut self) -> Result<()> {
+        let source = self
             .source
+            .as_ref()
+            .ok_or_else(|| anyhow!("No source file provided"))?;
+
+        let filename = source
             .file_name()
-            .ok_or_else(|| Error::Build("Missing file name".into()))?
+            .ok_or_else(|| anyhow!("Missing file name"))?
             .to_str()
-            .ok_or_else(|| Error::Build("Invalid unicode in file name".into()))?;
+            .ok_or_else(|| anyhow!("Invalid unicode in file name"))?;
+
         if !filename.ends_with(".bpf.c") {
-            return Err(Error::Build(format!(
-                "Source file={} does not have .bpf.c suffix",
-                self.source.display()
-            )));
+            return Err(anyhow!(
+                "Source `{}` does not have .bpf.c suffix",
+                source.display()
+            ));
         }
 
-        // Safe to unwrap b/c we already checked suffix
-        let name = filename.split('.').next().unwrap();
-        let dir = tempdir().map_err(|e| Error::Build(e.to_string()))?;
-        let objfile = dir.path().join(format!("{}.o", name));
+        if self.obj.is_none() {
+            let name = filename.split('.').next().unwrap();
+            let dir = tempdir().context("failed to create temporary directory")?;
+            let objfile = dir.path().join(format!("{name}.o"));
+            self.obj = Some(objfile);
+            // Hold onto tempdir so that it doesn't get deleted early
+            self.dir = Some(dir);
+        }
 
         build::build_single(
             self.debug,
-            &self.source,
-            &objfile,
+            source,
+            // Unwrap is safe here since we guarantee that obj.is_some() above
+            self.obj.as_ref().unwrap(),
             self.clang.as_ref(),
             self.skip_clang_version_check,
-            &self.clang_args,
+            self.clang_args.clone(),
         )
-        .map_err(|e| Error::Build(e.to_string()))?;
+        .with_context(|| format!("failed to build `{}`", source.display()))?;
+
+        Ok(())
+    }
+
+    // Generate a skeleton at path `output` without building BPF programs.
+    //
+    // [`SkeletonBuilder::obj`] must be set for this to succeed.
+    pub fn generate<P: AsRef<Path>>(&mut self, output: P) -> Result<()> {
+        let objfile = self.obj.as_ref().ok_or_else(|| anyhow!("No object file"))?;
 
         gen::gen_single(
             self.debug,
-            &objfile,
+            objfile,
             gen::OutputDest::File(output.as_ref()),
             Some(&self.rustfmt),
         )
-        .map_err(|e| Error::Generate(e.to_string()))?;
+        .with_context(|| format!("failed to generate `{}`", objfile.display()))?;
 
         Ok(())
     }
