@@ -1,9 +1,14 @@
+#![allow(clippy::let_unit_value)]
+#![warn(clippy::absolute_paths)]
+
+use std::ffi::OsString;
 use std::path::PathBuf;
 
 use anyhow::Result;
-use structopt::StructOpt;
+use clap::Args;
+use clap::Parser;
+use clap::Subcommand;
 
-mod btf;
 #[doc(hidden)]
 mod build;
 mod gen;
@@ -11,10 +16,15 @@ mod make;
 mod metadata;
 
 #[doc(hidden)]
-#[derive(Debug, StructOpt)]
+#[derive(Debug, Parser)]
+#[command(version, about)]
+#[command(propagate_version = true)]
 struct Opt {
-    #[structopt(subcommand)]
+    #[command(subcommand)]
     wrapper: Wrapper,
+    /// Enable debug output.
+    #[clap(short, long, global = true)]
+    debug: bool,
 }
 
 // cargo invokes subcommands with the first argument as
@@ -28,42 +38,47 @@ struct Opt {
 //
 // so we must have a dummy subcommand here to eat the arg.
 #[doc(hidden)]
-#[derive(Debug, StructOpt)]
+#[derive(Debug, Subcommand)]
 enum Wrapper {
+    #[command(subcommand)]
     Libbpf(Command),
 }
 
-#[doc(hidden)]
-#[derive(Debug, StructOpt)]
-#[structopt(verbatim_doc_comment)]
-///
+/// A grouping of clang specific options.
+#[derive(Debug, Args)]
+pub struct ClangOpts {
+    /// Path to clang binary
+    #[arg(long, value_parser)]
+    clang_path: Option<PathBuf>,
+    /// Additional arguments to pass to `clang`.
+    #[arg(long, value_parser)]
+    clang_args: Vec<OsString>,
+    /// Skip clang version checks
+    #[arg(long)]
+    skip_clang_version_checks: bool,
+}
+
 /// cargo-libbpf is a cargo subcommand that helps develop and build eBPF (BPF) programs.
+#[doc(hidden)]
+#[derive(Debug, Subcommand)]
 enum Command {
     /// Build bpf programs
     Build {
-        #[structopt(short, long)]
-        debug: bool,
-        #[structopt(long, parse(from_os_str))]
+        #[arg(long, value_parser)]
         /// Path to top level Cargo.toml
         manifest_path: Option<PathBuf>,
-        #[structopt(long, parse(from_os_str))]
-        /// Path to clang binary
-        clang_path: Option<PathBuf>,
-        #[structopt(long)]
-        /// Skip clang version checks
-        skip_clang_version_checks: bool,
+        #[command(flatten)]
+        clang_opts: ClangOpts,
     },
     /// Generate skeleton files
     Gen {
-        #[structopt(short, long)]
-        debug: bool,
-        #[structopt(long, parse(from_os_str))]
+        #[arg(long, value_parser)]
         /// Path to top level Cargo.toml
         manifest_path: Option<PathBuf>,
-        #[structopt(long, parse(from_os_str))]
+        #[arg(long, value_parser)]
         /// Path to rustfmt binary
         rustfmt_path: Option<PathBuf>,
-        #[structopt(long, parse(from_os_str))]
+        #[arg(long, value_parser)]
         /// Generate skeleton for the specified object file and print results to stdout
         ///
         /// When specified, skeletons for the rest of the project will not be generated
@@ -71,25 +86,19 @@ enum Command {
     },
     /// Build project
     Make {
-        #[structopt(short, long)]
-        debug: bool,
-        #[structopt(long, parse(from_os_str))]
+        #[arg(long, value_parser)]
         /// Path to top level Cargo.toml
         manifest_path: Option<PathBuf>,
-        #[structopt(long, parse(from_os_str))]
-        /// Path to clang binary
-        clang_path: Option<PathBuf>,
-        #[structopt(long)]
-        /// Skip clang version checks
-        skip_clang_version_checks: bool,
-        #[structopt(short, long)]
+        #[command(flatten)]
+        clang_opts: ClangOpts,
+        #[arg(short, long)]
         /// Quiet output
         quiet: bool,
         /// Arguments to pass to `cargo build`
         ///
         /// Example: cargo libbpf build -- --package mypackage
         cargo_build_args: Vec<String>,
-        #[structopt(long, parse(from_os_str))]
+        #[arg(long, value_parser)]
         /// Path to rustfmt binary
         rustfmt_path: Option<PathBuf>,
     },
@@ -97,23 +106,27 @@ enum Command {
 
 #[doc(hidden)]
 fn main() -> Result<()> {
-    let opts = Opt::from_args();
+    let opts = Opt::parse();
+    let Opt { wrapper, debug } = opts;
 
-    match opts.wrapper {
+    match wrapper {
         Wrapper::Libbpf(cmd) => match cmd {
             Command::Build {
-                debug,
                 manifest_path,
-                clang_path,
-                skip_clang_version_checks,
+                clang_opts:
+                    ClangOpts {
+                        clang_path,
+                        clang_args,
+                        skip_clang_version_checks,
+                    },
             } => build::build(
                 debug,
                 manifest_path.as_ref(),
                 clang_path.as_ref(),
+                clang_args,
                 skip_clang_version_checks,
             ),
             Command::Gen {
-                debug,
                 manifest_path,
                 rustfmt_path,
                 object,
@@ -124,10 +137,13 @@ fn main() -> Result<()> {
                 object.as_ref(),
             ),
             Command::Make {
-                debug,
                 manifest_path,
-                clang_path,
-                skip_clang_version_checks,
+                clang_opts:
+                    ClangOpts {
+                        clang_path,
+                        clang_args,
+                        skip_clang_version_checks,
+                    },
                 quiet,
                 cargo_build_args,
                 rustfmt_path,
@@ -135,6 +151,7 @@ fn main() -> Result<()> {
                 debug,
                 manifest_path.as_ref(),
                 clang_path.as_ref(),
+                clang_args,
                 skip_clang_version_checks,
                 quiet,
                 cargo_build_args,
